from abc import ABC, abstractmethod

from kubernetes.client.rest import ApiException
import pytest
import random
import yaml

from si_tests import logger
from si_tests import settings
from si_tests.managers.kaas_manager import Manager
from si_tests.utils import waiters

LOG = logger.logger


class AbstractModule(ABC):

    @abstractmethod
    def get_config(self,  *args, **kwargs):
        raise NotImplementedError("Abstract method")

    @abstractmethod
    def list_configs(self, *args, **kwargs):
        raise NotImplementedError("Abstract method")

    @abstractmethod
    def create(self, *args, **kwargs):
        raise NotImplementedError("Abstract method")

    @abstractmethod
    def delete(self, *args, **kwargs):
        raise NotImplementedError("Abstract method")


class BaseModule(AbstractModule):

    def __init__(self, cluster, ns):
        self.cluster = cluster
        self.ns = ns

    def get_config(self, name):
        result = self.list_configs(name_prefix=name)
        if not result:
            raise Exception(f"HostOSConfiguration '{name}' not found")
        if len(result) > 1:
            raise Exception(f"Found more than one HostOSConfiguration with name '{name}' "
                            f"in the namespace {self.cluster.namespace}")
        return result[0]

    def list_configs(self, name_prefix=None):
        return self.cluster.get_hostosconfigurations(name_prefix=name_prefix)

    def create(self, name, configs, order, machines_type='control'):
        """Create HostOSConfiguration"""
        match_labels = self.get_match_labels(machines_type)

        hostoscfg_data = {
            "apiVersion": "kaas.mirantis.com/v1alpha1",
            "kind": "HostOSConfiguration",
            "metadata": {
                "name": name,
                "namespace": self.cluster.namespace,
            },
            "spec": {
                "order": order,
                "configs": configs,
                "machineSelector": {
                    "matchLabels": match_labels,
                }
            }
        }
        hostoscfg = self.cluster.create_hostosconfiguration_raw(hostoscfg_data)
        return hostoscfg

    def patch(self, name, configs, machines_type='control'):
        """Create HostOSConfiguration"""
        match_labels = self.get_match_labels(machines_type)

        hostoscfg_data = {
            "spec": {
                "configs": configs,
                "machineSelector": {
                    "matchLabels": match_labels,
                }
            }
        }
        obj = self.get_config(name)
        hostoscfg = obj.patch(body=hostoscfg_data)
        return hostoscfg

    def delete(self, name, timeout=60, interval=5):
        """Delete HostOSConfiguration"""
        existing_configs = self.list_configs(name_prefix=name)
        for existing_config in existing_configs:
            LOG.info(f"Delete HostOSConfiguration '{existing_config.name}'")
            existing_config.delete()
        timeout_msg = f"HostOSConfiguration {name} was not deleted in {timeout}s"
        waiters.wait(lambda: not bool(self.list_configs(name_prefix=name)),
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=timeout_msg)

    def get_match_labels(self, machine_type='control'):
        if self.cluster.is_management or self.cluster.is_regional:
            # Management cluster have only 'control' node type
            match_labels_by_types = {
                'control': {
                    "cluster.sigs.k8s.io/control-plane": "true",
                },
                'worker': {
                    "kaas.mirantis.com/provider": "baremetal",
                },
            }
            assert machine_type in match_labels_by_types, (
                f"Machine type '{machine_type}' is not supported, please update the test")
            match_labels = match_labels_by_types.get(machine_type)
        else:
            match_labels_by_types = {
                'control': {
                    "hostlabel.bm.kaas.mirantis.com/controlplane": "controlplane",
                },
                'worker': {
                    "hostlabel.bm.kaas.mirantis.com/worker": "worker",
                },
                'storage': {
                    "hostlabel.bm.kaas.mirantis.com/storage": "storage",
                },
            }
            assert machine_type in match_labels_by_types, (
                f"Machine type '{machine_type}' is not supported, please update the test")
            match_labels = match_labels_by_types.get(machine_type)
        return match_labels

    def apply_new_config(self, name, configs, order, machines_type='control'):
        # Remember LCMMachines timestamps before creating HostOSConfiguration
        lcmmachines_timestamps_before = self.cluster.get_cluster_lcmmachines_timestamps()

        hostoscfg = self.create(name=name, configs=configs, order=order, machines_type=machines_type)

        # Wait for the selected Machines in the hostosconfiguration status
        # Check that machines from hostosconfiguration status field have labels used for machineSelector
        self.cluster.check.check_hostosconfig_machine_selector(hostoscfg)
        # Check that new items added into machineTypes in LCMCluster
        self.cluster.check.wait_lcmcluster_day2_machinetypes(hostoscfg)
        # Check that new items added into stateItems in LCMMachine
        self.cluster.check.wait_lcmmachine_day2_stateitems(hostoscfg, lcmmachines_timestamps_before)
        # Check the result on the machines
        machine_names = self.get_hostosconfig_machines_status(hostoscfg)
        self.cluster.check.check_machines(hostoscfg, machine_names)

    def apply_new_config_negative(self, name, configs, order, machines_type='control'):

        # Remember LCMMachines timestamps before creating HostOSConfiguration
        lcmmachines_timestamps_before = self.cluster.get_cluster_lcmmachines_timestamps()

        hostoscfg = self.create(name=name, configs=configs, order=order, machines_type=machines_type)

        # Wait for the selected Machines in the hostosconfiguration status
        # Check that machines from hostosconfiguration status field have labels used for machineSelector
        self.cluster.check.check_hostosconfig_machine_selector(hostoscfg)
        # Check that new items added into machineTypes in LCMCluster
        self.cluster.check.wait_lcmcluster_day2_machinetypes(hostoscfg)

        # Check that new items added into stateItems in LCMMachine but stateItemStatuses contain errors
        self.cluster.check.wait_lcmmachine_day2_stateitems(
            hostoscfg, lcmmachines_timestamps_before, expected_error_result=True)

    def apply_patch(self, name, configs, machines_type='control'):

        # Remember LCMMachines timestamps before creating HostOSConfiguration
        lcmmachines_timestamps_before = self.cluster.get_cluster_lcmmachines_timestamps()

        hostoscfg = self.patch(name=name, configs=configs, machines_type=machines_type)

        # Wait for the selected Machines in the hostosconfiguration status
        # Check that machines from hostosconfiguration status field have labels used for machineSelector
        self.cluster.check.check_hostosconfig_machine_selector(hostoscfg)
        # Check that new items added into machineTypes in LCMCluster
        self.cluster.check.wait_lcmcluster_day2_machinetypes(hostoscfg)
        # Check that new items added into stateItems in LCMMachine
        self.cluster.check.wait_lcmmachine_day2_stateitems(hostoscfg, lcmmachines_timestamps_before)
        # Check the result on the machines
        machine_names = self.get_hostosconfig_machines_status(hostoscfg)
        self.cluster.check.check_machines(hostoscfg, machine_names)


class SysctlModule(BaseModule):
    """Test 'sysctl' module"""
    base_order = 100

    def list_configs(self, *args, **kwargs):
        return super().list_configs(*args, **kwargs)

    def create(self, *args, **kwargs):
        return super().create(*args, **kwargs)

    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    def get_values_to_create(self):
        return {
            "kernel.sysrq": "100" + str(random.randint(100, 199)),  # string value
            "net.ipv4.ip_forward": 1,  # integer value
        }

    def get_new_values_to_create(self):
        return {
            "vm.swappiness": 33,
        }

    def get_new_values_to_update(self):
        return {
            "kernel.sysrq": "100" + str(random.randint(500, 599)),  # string value
            "vm.swappiness": 31,  # override some old value
            "vm.dirty_ratio": 45,  # add new value
        }

    def get_bad_values(self):
        return {
            "kernel.aaa.bbb": str(random.randint(1000, 1999)),
        }

    def get_secret_values_to_create(self):
        return {
            "kernel.sysrq": "100" + str(random.randint(700, 799)),  # string value
            "vm.swappiness": 10,
        }

    def check_machines(self, hostoscfg, chosen_machine_names):
        self.cluster.check.check_day2_sysctl_module_results(hostoscfg, chosen_machine_names)


class PackageModule(BaseModule):
    """Test 'package' module"""
    base_order = 200

    def list_configs(self, *args, **kwargs):
        return super().list_configs(*args, **kwargs)

    def create(self, *args, **kwargs):
        return super().create(*args, **kwargs)

    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    def get_values_to_create(self):
        return {
            "elinks": "present",
        }

    def get_new_values_to_create(self):
        return {
            "lynx": "present",
        }

    def get_new_values_to_update(self):
        return {
            "lynx": "absent",  # override some old value
            "tshark": "present",  # add new value
        }

    def get_bad_values(self):
        return {
            "system-integration-2024-" + str(random.randint(10, 99)): "present",
        }

    def get_secret_values_to_create(self):
        return {
            "dnstop": "present",
        }

    def check_machines(self, hostoscfg, chosen_machine_names):
        self.cluster.check.check_day2_package_module_results(hostoscfg, chosen_machine_names)


def _get_testobject_class(module_name):
    _modules = {
        'sysctl': SysctlModule,
        'package': PackageModule,
    }
    assert module_name in _modules, (f"Undefined module name {module_name}")
    testobject_class = _modules[module_name]
    return testobject_class


def generate_module_data():
    kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    cluster_name = settings.TARGET_CLUSTER
    cluster_ns = settings.TARGET_NAMESPACE
    ns = kaas_manager.get_namespace(namespace=cluster_ns)
    cluster = ns.get_cluster(cluster_name)

    # Collect modules data from hostosconfigurationmodules objects
    hostoscfgmodules_list = kaas_manager.get_hostosconfigurationmodules()
    test_modules = {}
    for hostoscfgmodules in hostoscfgmodules_list:
        hostoscfgmodules_data = hostoscfgmodules.data
        # Collect modules from spec
        hostoscfg_modules_spec = hostoscfgmodules_data.get('spec', {}).get('modules', [])
        for hostoscfg_module in hostoscfg_modules_spec:
            hostoscfg_module_name = hostoscfg_module.get("name")
            hostoscfg_module_version = hostoscfg_module.get("version")
            hostoscfg_module_key = f"{hostoscfg_module_name}-{hostoscfg_module_version}"
            assert hostoscfg_module_key not in test_modules.keys(), (
                f"Duplicate definition of the module '{hostoscfg_module_name}:{hostoscfg_module_version}' "
                f"in both HostOSConfigurationModule objects '{hostoscfgmodules.name}' and "
                f"'{test_modules[hostoscfg_module_key]['object_ref']}'")

            testobject_class = _get_testobject_class(hostoscfg_module_name)
            testobject = testobject_class(cluster, ns)

            test_modules[hostoscfg_module_key] = {
                'object_ref': hostoscfgmodules.name,
                'key': hostoscfg_module_key,
                'spec': hostoscfg_module,
                'testobject': testobject,
                'failed_test': '',
            }

        # Collect modules from status
        hostoscfg_modules_status = hostoscfgmodules_data.get('status', {}).get('modules', [])
        for hostoscfg_module in hostoscfg_modules_status:
            hostoscfg_module_name = hostoscfg_module.get("name")
            hostoscfg_module_version = hostoscfg_module.get("version")
            hostoscfg_module_key = f"{hostoscfg_module_name}-{hostoscfg_module_version}"
            test_modules[hostoscfg_module_key]['status'] = hostoscfg_module

    return test_modules.values()


modules_data = list(generate_module_data())


@pytest.fixture(scope='function', params=modules_data,
                ids=[f"MODULE={x['key']}" for x in modules_data])
def module_data(request, func_name):
    _data = request.param
    if _data['failed_test']:
        msg = (f"Skip '{func_name}' because previous test for the module {_data['key']} has been failed")
        LOG.info(msg)
        pytest.skip(msg)

    LOG.info(f"Run test {func_name} for module {_data['key']}")
    yield _data
    # Check the result of the current step
    test_passed = (hasattr(request.node, 'rep_call') and
                   request.node.rep_call.passed)
    if not test_passed:
        _data['failed_test'] = func_name


def test_010_host_os_configuration_si(module_data, kaas_manager):
    """Test HostOSConfigurations and HostOSConfigurationModules
        8. TODO(ddmitriev): Test applying 'values' with the same module name but with higher module version

    """
    module_state = module_data['status'].get('state')
    module_error = module_data['status'].get('error')
    assert module_state == 'available', (
        f"HostOSConfigurationModule '{module_data['key']}' is in wrong state '{module_state}': {module_error}\n"
        f"Module spec:\n{yaml.safe_dump(module_data['spec'])}\n"
        f"Module status:\n{yaml.safe_dump(module_data['status'])}")


def test_020_host_os_configuration_si_negative_conflict_options(module_data, kaas_manager):
    """Negative: create HostOSConfiguration with 'values' and 'secretValues' at the same time should not be allowed

    Expected result: admission controller does not accept such set of options, and return HTTP status = 400
    """

    hostoscfg_module_name = module_data['spec']['name']
    hostoscfg_module_version = module_data['spec']['version']
    testobject = module_data['testobject']
    hostoscfg_name = f"test-020-negative-{module_data['key']}-{testobject.cluster.name}"
    # Remove existing HostOSConfiguration objects with such names
    testobject.delete(name=hostoscfg_name)

    hostoscfg_configs = [
        {
            "module": hostoscfg_module_name,
            "moduleVersion": hostoscfg_module_version,
            "values": testobject.get_values_to_create(),
            "secretValues": testobject.get_secret_values_to_create(),
        },
    ]
    hostoscfg_order = testobject.base_order + 20

    with pytest.raises(ApiException) as exc_info:
        # Try to create hoc object, expecting result status 400
        testobject.create(name=hostoscfg_name, configs=hostoscfg_configs,
                          order=hostoscfg_order, machines_type='control')
        pytest.fail(f"Expected result with 'Bad Request', but object '{hostoscfg_name}' has been succesfully created")
    assert exc_info.value.status == 400 and exc_info.value.reason == 'Bad Request', (
        f"Expected reason 'Bad Request' when module config contains both 'values' and 'secretValues', "
        f"but got:\n{exc_info.value.body}")
    LOG.info(f"Got the following response: {exc_info.value.body}")


def test_021_host_os_configuration_si_negative_bad_values(module_data, kaas_manager):
    """Negative: create HostOSConfiguration with bad values

    Expected result: HostOSConfiguration is created, but bad values cause errors in LCMMachines stateItemStatuses
    """

    hostoscfg_module_name = module_data['spec']['name']
    hostoscfg_module_version = module_data['spec']['version']
    testobject = module_data['testobject']
    hostoscfg_name = f"test-021-negative-{module_data['key']}-{testobject.cluster.name}"
    # Remove existing HostOSConfiguration objects with such names
    testobject.delete(name=hostoscfg_name)
    # 1. Create HostOSConfiguration with bad values and ensure
    #    that all related LCMMachines got errors for the related stateItemStatuses
    hostoscfg_configs = [
        {
            "module": hostoscfg_module_name,
            "moduleVersion": hostoscfg_module_version,
            "values": testobject.get_bad_values(),
        },
    ]
    hostoscfg_order = testobject.base_order + 21

    try:
        testobject.apply_new_config(name=hostoscfg_name, configs=hostoscfg_configs,
                                    order=hostoscfg_order, machines_type='control')
    finally:
        # Remove created HostOSConfiguration object
        testobject.delete(name=hostoscfg_name)


def test_030_host_os_configuration_si_first_config(module_data, kaas_manager):
    """Test applying 'values' in a new HostOSConfiguration"""

    hostoscfg_module_name = module_data['spec']['name']
    hostoscfg_module_version = module_data['spec']['version']
    testobject = module_data['testobject']
    hostoscfg_name = f"test-030-first-config-{module_data['key']}-{testobject.cluster.name}"
    # Remove existing HostOSConfiguration objects with such names
    testobject.delete(name=hostoscfg_name)

    hostoscfg_configs = [
        {
            "module": hostoscfg_module_name,
            "moduleVersion": hostoscfg_module_version,
            "values": testobject.get_values_to_create(),
        },
    ]
    hostoscfg_order = testobject.base_order + 30

    # Create a HostOSConfiguration on 'control' nodes on Management or Child cluster
    testobject.apply_new_config(name=hostoscfg_name, configs=hostoscfg_configs,
                                order=hostoscfg_order, machines_type='control')


def test_040_host_os_configuration_si_second_config(module_data, kaas_manager):
    """Test applying new HostOSConfiguration object with 'values' for already existing parameters."""
    hostoscfg_module_name = module_data['spec']['name']
    hostoscfg_module_version = module_data['spec']['version']
    testobject = module_data['testobject']
    hostoscfg_name = f"test-040-second-config-{module_data['key']}-{testobject.cluster.name}"
    # Remove existing HostOSConfiguration objects with such names
    testobject.delete(name=hostoscfg_name)

    hostoscfg_configs = [
        {
            "module": hostoscfg_module_name,
            "moduleVersion": hostoscfg_module_version,
            "values": testobject.get_new_values_to_create(),
        },
    ]
    hostoscfg_order = testobject.base_order + 40

    # TODO(ddmitriev): also change the machine_type for this object from 'control' to 'storage' for child
    testobject.apply_new_config(name=hostoscfg_name, configs=hostoscfg_configs,
                                order=hostoscfg_order, machines_type='control')


def test_050_host_os_configuration_si_update_first_config(module_data, kaas_manager):
    """Test that new 'values' added to an existing HostOSConfiguration object are applied on the Machines"""
    hostoscfg_module_name = module_data['spec']['name']
    hostoscfg_module_version = module_data['spec']['version']
    testobject = module_data['testobject']

    hostoscfg_configs = [
        {
            "module": hostoscfg_module_name,
            "moduleVersion": hostoscfg_module_version,
            "values": testobject.get_new_values_to_update(),
        },
    ]

    # Use existing HostOSConfiguration from previous test to update it's values
    hostoscfg_name = f"test-030-first-config-{module_data['key']}-{testobject.cluster.name}"

    # TODO(ddmitriev): also change the machine_type for this object from 'control' to 'worker' for child
    testobject.apply_patch(name=hostoscfg_name, configs=hostoscfg_configs)


def test_060_host_os_configuration_si_ordered_values(module_data, kaas_manager):
    """Test overriding applied values with new values with a higher 'order' in a new HostOSConfigurations"""
    # TODO(ddmitriev): feature 'order' is in refactoring, need to wait for a stable documentation
    pass


def test_070_host_os_configuration_si_add_secret_values(module_data, kaas_manager):
    """Test applying 'secretValues' in a new HostOSConfiguration"""
    hostoscfg_module_name = module_data['spec']['name']
    hostoscfg_module_version = module_data['spec']['version']
    testobject = module_data['testobject']
    hostoscfg_name = f"test-070-add-secret-values-{module_data['key']}-{testobject.cluster.name}"
    # Remove existing HostOSConfiguration objects with such names
    testobject.delete(name=hostoscfg_name)

    # Remove existing Secret objects with such name
    testobject.ns.delete_secret(hostoscfg_name)

    secret_data = testobject.get_secret_values_to_create()
    testobject.ns.create_secret(name=hostoscfg_name, data=secret_data)

    hostoscfg_configs = [
        {
            "module": hostoscfg_module_name,
            "moduleVersion": hostoscfg_module_version,
            "secretValues": {
                # Secret object name
                "name": hostoscfg_name,
                "namespace": testobject.ns.name,
            },
        },
    ]
    hostoscfg_order = testobject.base_order + 70

    testobject.apply_new_config(name=hostoscfg_name, configs=hostoscfg_configs,
                                order=hostoscfg_order, machines_type='control')


def test_080_host_os_configuration_si_mixed_modules_configuration(kaas_manager):
    """Add HostOSConfiguration with values for all existing modules in a single object"""
    # for module_data in modules_data:
    #     testobject = module_data['testobject']
    pass
    # TODO(ddmitriev): generate data for all supported modules and create hoc object


def test_999_applied_host_os_configurations_si(kaas_manager, show_step):
    """Ensure that nodes still have the expected configurations applied with HostOSConfigurations

    Test should be used after MCC LCM operations to ensure that configuration options are persistent

    Scenario:
        1. Collect all HostOSConfigurations for the Cluster
        2. For each LCMMachine, leave only those HostOSConfigurations, which were created after the LCMMachine
        3. Sort filtered HostOSConfigurations by 'creationTimestamp', from older to newer
        4. Collect values from sorted 'HostOSConfigurations' in a dict like {<module_name>: {<option_name>: <value>}},
           with overwriting older values with newer values
        5. Ensure that the collected values are actually present on the Machine
    """
    # cluster_name = settings.TARGET_CLUSTER
    # cluster_ns = settings.TARGET_NAMESPACE
    # ns = kaas_manager.get_namespace(namespace=cluster_ns)
    # cluster = ns.get_cluster(cluster_name)
    pass
    # TODO(ddmitriev): Ensure that current cluster Machines contain the applied configuration
    #                  from all HostOSConfiguration objects
    # 1. iterate over all HostOSConfiguration modules
    # 2. collect the parameters and values per each HostOSConfiguration module
    # 3. merge the collected parameters and values depending to the expected order (under discussion atm)
    # 4. check that the merged parameters and values match the actual configurations on the Machines
