#!/usr/bin/env python3
import argparse
import os
import sys
from collections import namedtuple
from tabulate import tabulate

sys.path.append(os.getcwd())
try:
    from si_tests.utils import packaging_version as version
    from si_tests import settings
    from si_tests.managers.kaas_manager import Manager
except ImportError:
    print("ImportError: Run the application from the si-tests directory or "
          "set the PYTHONPATH environment variable to directory which contains"
          " ./si_tests")
    sys.exit(1)

ReleaseUpgrade = namedtuple('ReleaseUpgrade',
                            'version,numeric_version,destination_cr_ver,upgrade_path_list,'
                            'upgrade_path_str,available_upgrades')
AvailableUpgrade = namedtuple('AvailableUpgrade',
                              'version,numeric_version,reboot_required,skip_maintenance')


def parse_args():
    parser = argparse.ArgumentParser(description="Release Upgrade Tool")
    parser.add_argument("--provider", choices=["openstack", "vsphere", "byo"], required=True,
                        help="Specify the provider (openstack, vsphere)")
    parser.add_argument("--output", choices=["display", "generate"], required=True,
                        help="Specify command for display upgrade path or generate pairs)")
    return parser.parse_args()


def generate_mke_pairs(provider, output):
    if not settings.KUBECONFIG_PATH or not os.path.isfile(
            settings.KUBECONFIG_PATH):
        print("Please set KUBECONFIG environment variable whith"
              "the path to the kubeconfig file for KaaS management cluster")
        return 11

    kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    mgm_cluster = kaas_manager.get_mgmt_cluster()
    deployed_kaas_release_name = mgm_cluster.get_kaasrelease_version()
    kaas_release = kaas_manager.get_kaasrelease(deployed_kaas_release_name)
    kaas_release = kaas_release.data
    supported_clusterreleases = kaas_release['spec']['supportedClusterReleases']
    releases = [item for item in supported_clusterreleases if provider in item["providers"]["supported"]]
    mke_releases = [item for item in releases if item["name"].startswith("mke")]
    mke_releases_versions = {mke_release['version']: mke_release['name'] for mke_release in mke_releases}
    mke_releases_latest_version = sorted(mke_releases_versions.keys(), key=version.parse)[-1]
    #########################################################
    # WA for vsphere provider
    if mgm_cluster.workaround.prodx_26477() and provider != "byo":
        if provider == 'vsphere':
            target_cluster_release_mcc = "mke-14-1-0-rc-3-6-6"
            excluded_releases = ['mke-16-0-0-rc-3-7-1', 'mke-16-0-0-3-7-1']
        elif provider == "openstack":
            target_cluster_release_mcc = kaas_release['spec']['clusterRelease']
            excluded_releases = ['mke-14-1-0-rc-3-6-6', 'mke-14-1-0-3-6-6']
    else:
        target_cluster_release_mcc = mke_releases_versions[mke_releases_latest_version]
        excluded_releases = []
    #########################################################

    def get_release_by_cr(cr_version):
        return next((item for item in supported_clusterreleases if item["name"] == cr_version), None)

    def get_cr_by_numeric(numeric_ver):
        return next((item["name"] for item in supported_clusterreleases if item["version"] == numeric_ver), None)

    def get_upgrade_path(item, target_ver, excluded_r, rr=False, sm=False):
        nonlocal upgrade_path_res
        for a in item.get('availableUpgrades', []):
            tmp_cr_ver = get_cr_by_numeric(a['version'])
            rr = a.get("rebootRequired", False)
            sm = a.get("skipMaintenance", False)

            if tmp_cr_ver in excluded_r:
                continue

            upgrade_path_res.append(AvailableUpgrade(
                version=tmp_cr_ver,
                numeric_version=a['version'],
                reboot_required=rr,
                skip_maintenance=sm
            ))

            if tmp_cr_ver == target_ver:
                return upgrade_path_res
            else:
                release = get_release_by_cr(tmp_cr_ver)
                get_upgrade_path(release, target_ver, excluded_releases, rr=rr, sm=sm)
        return upgrade_path_res

    res_print = []
    result = []
    upgrade_path_res = []

    def collect_path(releases, target_cr, excluded_release):
        for item in releases:
            nonlocal upgrade_path_res
            upgrade_path_res = list()
            upgrade_path = get_upgrade_path(item, target_cr, excluded_release) or []
            version = item.get("name")
            numeric_version = item.get("version")
            destination_cr_ver = target_cr
            upgrade_path_list = [x.version for x in upgrade_path]
            upgrade_path_str = " -> ".join(upgrade_path_list) if upgrade_path_list else "------"
            available_upgrades = upgrade_path
            if version not in excluded_releases:
                result.append(
                    ReleaseUpgrade(
                        version=version,
                        numeric_version=numeric_version,
                        destination_cr_ver=destination_cr_ver,
                        upgrade_path_list=upgrade_path_list,
                        upgrade_path_str=upgrade_path_str,
                        available_upgrades=available_upgrades
                    )
                )

    collect_path(mke_releases, target_cluster_release_mcc, excluded_releases)

    if output == 'display':
        for r in result:
            res_print.append((
                r.version,
                r.destination_cr_ver,
                " -> ".join(["{} ({}){rr}{sm}".format(x.version, x.numeric_version,
                                                      rr=" RR" if x.reboot_required else "",
                                                      sm=" SM" if x.skip_maintenance else "") for x in
                             r.available_upgrades])
            ))
        print('RR - patchrelease, SM - skip maintenance')
        print('\n' + tabulate(res_print, tablefmt="presto", headers=["From", "To", "Path"]))

    elif output == 'generate':
        formatted_output = []
        for r in result:
            formatted_output.append(
                f"{r.version},{r.destination_cr_ver}"
            )

        print('\n'.join(formatted_output))


args = parse_args()
generate_mke_pairs(args.provider, args.output)
